    /*********************************************************************\
    *  Copyright (c) 1991 by Wen-King Su (wen-king@vlsi.cs.caltech.edu)   *
    *  Copyright (c) 1993 by Phil Richards (pgr@prg.ox.ac.uk)             *
    *                                                                     *
    *  You may copy or modify this file in any manner you wish, provided  *
    *  that this notice is always included, and that you hold the author  *
    *  harmless for any loss or damage resulting from the installation or *
    *  use of this software.                                              *
    \*********************************************************************/

#include <common.h>
#include <local/local.h>
#include <remote/remote.h>

#ifndef MINDELAY
# define MINDELAY 1000
#endif

#ifndef MAXRECURSION
# define MAXRECURSION 512
#endif

#if defined(__STDC__) && __GNUC__ == 2
#define PRINTF_ATTRIB(x,y) __attribute__ ((format (printf, x, y)))
#else
#define PRINTF_ATTRIB(x,y)
#endif


/****************************************************************************
* These structures are used to implement a opendir/readdir mechanism similar
* to that of the normal opendir/reader mechanism in unix.
****************************************************************************/

typedef struct DDLIST {
	struct DDLIST *next;
	char          *path;
	RDIRENT  **dep_root;
	u_int       ref_cnt;
} DDLIST;

typedef struct RDIR {
	DDLIST   *ddp;
        RDIRENT **dep;
} RDIR;

typedef struct rdirent {
	u_long  rd_fileno;
	u_short rd_reclen;
	u_short rd_namlen;
	char     *rd_name;
} rdirent;

typedef struct iobuffers {
	FILE	*in;	 /* stream to read all command input from        */
	FILE	*out;	 /* stream to send all command output to         */
	FILE	*err;	 /* stream to send all command errors to         */
	FILE	*prompt; /* stream to send all prompt messages to        */
	FILE	*info;	 /* stream to send all informational messages to */
	FILE	*warn;	 /* stream to send all warning messages to       */
	FILE	*dbg;	 /* stream to send all debug messages to         */
} iobuffers;

/* externs from lib.c */
extern UBUF *client_interact
    __P((u_int cmd, u_long pos, u_int l1, u_char *p1, u_int l2, u_char *p2));
extern void print_comm_stats __P((FILE *out));
extern int init_client __P((char *host, int port, int myport));
extern void finish_client __P((void));
extern void client_done __P((void));
extern int client_trace;
extern int client_intr_state;
extern int client_intr_cnt;
extern int key_persists;
extern u_long target_delay;
extern u_long udp_sent_time;
extern u_int time_out;
extern int burst_max;
extern u_short client_buf_len;
extern u_short client_net_len;

/* externs from util.c */
extern char *env_host;
extern char *env_port;
extern char *env_myport;
extern char *env_dir;

extern void ffprintf __P((FILE *out, const char *fmt, ...))
	PRINTF_ATTRIB(2,3);

extern char *my_fgets __P((char *s, int n, FILE *stream));

extern char *util_abs_path __P((const char *s2));
extern char *util_getwd __P((char *p));
extern RDIRENT **get_dir_blk __P((char *path));
extern int util_download_main
    __P((char *path, char *fpath, int fd,
	 int cmd, u_long offset, u_long length));
extern int util_download
    __P((char *path, int fd, u_long offset, u_long length));
extern int util_upload __P((char *path, FILE *fp));
extern int util_grab_file
    __P((char *path, int fd, u_long offset, u_long length));
extern void util_get_env __P((void));
extern void env_client __P((void));

extern void util_flushdir __P((void));
extern void util_dirtydir __P((char *path));
extern RDIR *util_opendir __P((char *path));
extern int util_closedir __P((RDIR *rdirp));
extern rdirent *util_readdir __P((RDIR *rdirp));
extern void util_split_path __P((char *path, char **p1, char **p2, char **p3));
extern int util_stat __P((const char *path, struct stat *sbuf));
extern int util_cd __P((char *p));

extern void util_flushpro __P((void));
extern void util_dirtypro __P((char *path));
extern int util_both_pro __P((char *path, char **textp,
			      char **bmapp, u_int *bmaplen));

extern int util_process_file
    __P((char *path, int recursive, int depth,
	 int (*process_file)(char *name, struct stat *sbufp, int depth),
         int (*process_dir)(char *name, int depth, char **private_data),
         void (*tidy_dir)(char *name, int depth, char *data)));
extern int util_process_arglist __P((char *argv[], int (*procfn)(char *name)));
extern int util_print_protection __P((char *name));

extern void util_dirty_version __P((void));
extern int util_print_version __P((void));

extern void util_print_readme __P((void));

/*
** kludge to allow validate_operation to act upon the filename specified
** rather than the basename of the file; this is only really used by rmdir
*/
#define LITERAL_DIR             0x80000000
extern int validate_operation __P((char *name, u_long opmask));

#define UTIL_PRINT_PROTECTION	(0)
#define UTIL_PRINT_VERSION	(0)
#define UTIL_CD			(DIR_PRIV)
#define UTIL_DIR		(DIR_PRIV)	/* opendir/readdir/closedir */
#define UTIL_DOWNLOAD		(DIR_PRIV)
#define UTIL_STAT		(DIR_PRIV)
#define UTIL_UPLOAD		(DIR_ADD)
#define UTIL_PROCESS_FILE	(UTIL_DIR | UTIL_STAT)
#define UTIL_GLOB		(UTIL_DIR | UTIL_STAT)
#define UTIL_GRAB_FILE		(UTIL_DOWNLOAD | DIR_DEL)
#define UTIL_PROCESS_ARGLIST	(UTIL_GLOB)

/* externs from redirect.c */
extern int askprompt;
extern int datestamp;

extern iobuffers global_iobuffers;

#define STDIN		(global_iobuffers.in)
#define STDOUT		(global_iobuffers.out)
#define STDERR		(global_iobuffers.err)
#define STDPROMPT	(global_iobuffers.prompt)
#define STDINFO		(global_iobuffers.info)
#define STDWARN		(global_iobuffers.warn)
#define STDDBG		(global_iobuffers.dbg)

extern int dbug_flag;
extern char *pager_command;

/* externs from glob.c */
#ifdef __STDC__
typedef void VOIDDIR;
typedef void VOIDDIRENT;
#else
typedef char VOIDDIR;
typedef char VOIDDIRENT;
#endif
extern int glob_match __P((char *pattern, char *string));
extern int glob_path __P((char *pattern, char *(*names[])));
extern char *expand_tilde __P((char *path));
extern char **glob __P((char *path));
extern void free_glob __P((char **argv));
extern void set_glob_routines
   __P((VOIDDIR *(*dopen)(char *dirname), void (*dclose)(VOIDDIR *dirp),
	VOIDDIRENT *(*dread)(VOIDDIR *dirp), char *(*dgetname)(VOIDDIRENT *dp),
	int (*dstat)(const char *buf, struct stat *s)));
void local_glob_routines __P((void));
void remote_glob_routines __P((void));

extern int standalone;
