    /*********************************************************************\
    *  Copyright (c) 1991 by Wen-King Su (wen-king@vlsi.cs.caltech.edu)   *
    *                                                                     *
    *  You may copy or modify this file in any manner you wish, provided  *
    *  that this notice is always included, and that you hold the author  *
    *  harmless for any loss or damage resulting from the installation or *
    *  use of this software.                                              *
    \*********************************************************************/

#include "client.h"
#include "lock.h"

#ifndef NOLOCKING

static char key_string[sizeof(KEY_PREFIX)+32];

static char code_str[] =
    "0123456789:_ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

static void make_key_string(server_addr, server_port)
    u_long server_addr; 
    u_int server_port;
{
    u_long v1, v2;
    char *p;

    (void)strcpy(key_string, KEY_PREFIX);
    for (p = key_string; *p; p++)
	;
    v1 = server_addr;
    v2 = server_port;

    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6; v1 = v1 | (v2 << (32-3*6));
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p++ = code_str[v1 & 0x3f]; v1 >>= 6;
    *p   = 0;
}

#endif

/********************************************************************/
/******* For those systems that has flock function call *************/
/********************************************************************/
#ifdef USE_FLOCK

#include <sys/file.h>

int key_persists = 1;
static u_int lock_fd;
static u_int okey;

int client_get_key __P((void))
{
    if (flock(lock_fd, LOCK_EX) < 0)
    {
	perror("flock");
	exit(1);
    }
    if (read(lock_fd, (char*)&okey, sizeof(okey)) < 0)
    {
	perror("read");
	exit(1);
    }
    if (lseek(lock_fd, 0L, 0) < 0)
    {
	perror("seek");
	exit(1);
    }

    return(okey);
}

void client_put_key(key)
    u_int key;
{
    if (write(lock_fd, (char*)&key, sizeof(key)) < 0)
    {
	perror("write");
	exit(1);
    }
    if (lseek(lock_fd, 0L, 0) < 0)
    {
	perror("seek");
	exit(1);
    }
    if (flock(lock_fd, LOCK_UN) < 0)
    {
	perror("unflock");
	exit(1);
    }
}

void client_init_key(server_addr, server_port, key)
    u_long server_addr; 
    u_int server_port;
    u_int key;
{
    u_long omask;
    okey = key;

    make_key_string(server_addr, server_port);

    omask = umask(0);
    lock_fd = open(key_string, O_RDWR | O_CREAT, 0666);
    (void)umask(omask);
}

#endif
/********************************************************************/
/******* For those systems that has lockf function call *************/
/********************************************************************/
#ifdef USE_LOCKF

#include <unistd.h>

int key_persists = 1;
static u_int lock_fd;
static u_int okey;

int client_get_key __P((void))
{
    if (lockf(lock_fd, F_LOCK, sizeof(okey)) < 0)
    {
	perror("lockf");
	exit(1);
    }
    if (read(lock_fd, &okey, sizeof(okey)) < 0)
    {
	perror("readlk");
	exit(1);
    }
    if (lseek(lock_fd, 0L, 0) < 0)
    {
	perror("seek");
	exit(1);
    }

    return(okey);
}

void client_put_key(key)
    u_int key;
{
    if (write(lock_fd, &key, sizeof(key)) < 0)
    {
	perror("write");
	exit(1);
    }
    if (lseek(lock_fd, 0L, 0) < 0)
    {
	perror("seek");
	exit(1);
    }
    if (lockf(lock_fd, F_ULOCK, sizeof(key)) < 0)
    {
	perror("unlockf");
	exit(1);
    }
}

void client_init_key(server_addr, server_port, key)
    u_long server_addr; 
    u_int server_port;
    u_int key;
{
    u_long omask;
    okey = key;

    make_key_string(server_addr, server_port);

    omask = umask(0);
    lock_fd = open(key_string, O_RDWR | O_CREAT, 0666);
    (void)umask(omask);
}

#endif
/********************************************************************/
/******* For those systems that has SysV shared memory + lockf ******/
/********************************************************************/
#ifdef USE_SHAREMEM_AND_LOCKF

#include <unistd.h>
#include <sys/ipc.h>
extern char *shmat();

int key_persists = 0;
static u_int *share_key;
static u_int lock_fd;

int client_get_key __P((void))
{
    if (lockf(lock_fd, F_LOCK, 2) < 0)
    {
	perror("lockf");
	exit(1);
    }
    return(*share_key);
}

void client_put_key(key)
    u_int key;
{
    *share_key = key;
    if (lockf(lock_fd, F_ULOCK, 2) < 0)
    {
	perror("unlockf");
	exit(1);
    }
}

void client_init_key(server_addr, server_port, key)
    u_long  server_addr; 
    u_int server_port;
    u_int key;
{
    u_long omask;
    key_t lock_key;
    int   lock_shm;

    make_key_string(server_addr, server_port);

    omask = umask(0);
    lock_fd = open(key_string, O_RDWR | O_CREAT, 0666);
    (void)umask(omask);

    if ((lock_key = ftok(key_string, 3432)) < 0)
    {
	perror("ftok");
	exit(1);
    }
    if ((lock_shm = shmget(lock_key, sizeof(short), IPC_CREAT | 0666)) < 0)
    {
	perror("shmget");
	exit(1);
    }
    if (!(share_key = (u_int *)shmat(lock_shm, 0, 0)))
    {
	perror("shmat");
	exit(1);
    }
}

#endif
/********************************************************************/
/******* For those who does not want to use locking *****************/
/********************************************************************/
#ifdef NOLOCKING

int key_persists = 0;
static u_int okey;

int client_get_key __P((void))
{
    return (int)okey;
}

void client_put_key(key)
    u_int key;
{
    okey = key;
}

void client_init_key(server_addr, server_port, key)
    u_long server_addr; 
    u_int server_port;
    u_int key;
{
    okey = key;
}

#endif
/********************************************************************/
/********************************************************************/
/********************************************************************/
